import logging

from typing import Type

from .filehandlers import BaseFileHandler, ResourceNotFoundError
from .keyring import Keyring
from .minivault import MiniVault
from .vaultflags import VaultFlags
from .vault import VarVault
from .utils import create_mini_vault_from_file


def create_vault(varvault_keyring: Type[Keyring],
                 varvault_vault_name: str,
                 *flags: VaultFlags,
                 varvault_filehandler_to: BaseFileHandler = None,
                 varvault_specific_logger: logging.Logger = None,
                 **extra_keys) -> VarVault:
    """
    Factory-function to help create a Vault-object instead of creating it manually (which is still possible).

    :param varvault_keyring: Describes which keys belong to this Vault.
    :param varvault_vault_name: Used to name the vault and the logger for writing information and debug messages.
    :param varvault_filehandler_to: Optional filehandler for a file to write the arguments in the vault to.
    :param flags: Optional argument for defining VaultFlags for this Vault.
     Note that any global VaultFlag will be overridden by VaultFlags defined in a vault-decorator.
    :param varvault_specific_logger: Optional argument for defining your own logger object if you want to use a specific logger rather than varvault's own logger.
    :param extra_keys: Extra keys as a dict to write to. These keys can be defined during runtime, which can sometimes be necessary.
     It is recommended to use pre-determined keys (e.g. constants), but sometimes being more flexible can be useful.
    :return: Vault object based on input to this function.
    """
    if not VaultFlags.flag_is_set(VaultFlags.remove_existing_log_file(), *flags):
        flags = (VaultFlags.remove_existing_log_file(), *flags)
    if varvault_filehandler_to:
        varvault_filehandler_to.live_update = VaultFlags.flag_is_set(VaultFlags.live_update(), *flags)
    return VarVault(varvault_keyring, varvault_vault_name, *flags,
                    varvault_filehandler_from=varvault_filehandler_to,
                    varvault_filehandler_to=varvault_filehandler_to,
                    varvault_specific_logger=varvault_specific_logger,
                    **extra_keys)


def from_vault(varvault_keyring: Type[Keyring],
               varvault_vault_name: str,
               varvault_filehandler_from: BaseFileHandler,
               *flags: VaultFlags,
               varvault_filehandler_to: BaseFileHandler = None,
               varvault_specific_logger: logging.Logger = None,
               **extra_keys) -> VarVault:
    """
    Factory-function to help create a Vault-object from an existing vault file.

    :param varvault_keyring: Describes which keys belong to this Vault.
    :param varvault_vault_name: Used to name the vault and the logger for writing information and debug messages.
    :param varvault_filehandler_from: Filehandler for a file to load variables from.
    :param flags: A set of VaultFlags to tweak the behavior of the vault.
    :param varvault_filehandler_to: Optional filehandler to write the variables to. This can be a separate file to the file we read from. If left as None, this will be set to 'varvault_filehandler_from'.
    :param varvault_specific_logger: Optional argument for defining your own logger object if you want to use a specific logger rather than varvault's own logger.
    This can be useful when changing the keyring without it being a real issue.
    :param extra_keys: Extra keys as a dict to write to. These keys can be defined during runtime, which can sometimes be necessary.
    It is recommended to use pre-determined keys (e.g. constants), but sometimes being more flexible can be useful.
    :return: A vault based on vault_filename_from and keyring.
    """
    ignore_keys_not_in_keyring = VaultFlags.flag_is_set(VaultFlags.vault_is_read_only(), *flags) or VaultFlags.flag_is_set(VaultFlags.ignore_keys_not_in_keyring(), *flags)
    varvault_filehandler_from.live_update = VaultFlags.flag_is_set(VaultFlags.live_update(), *flags)
    varvault_filehandler_from.vault_is_read_only = VaultFlags.flag_is_set(VaultFlags.vault_is_read_only(), *flags)

    if not varvault_filehandler_to:
        varvault_filehandler_to = varvault_filehandler_from

    keys_not_in_keyring = _check_for_keys_not_in_keyring(varvault_keyring, varvault_filehandler_from, ignore_keys_not_in_keyring, *flags, **extra_keys)
    if varvault_filehandler_from.exists():
        mini = create_mini_vault_from_file(varvault_filehandler_from, varvault_keyring, **extra_keys)
    elif not varvault_filehandler_from.exists() and VaultFlags.flag_is_set(VaultFlags.live_update(), *flags):
        mini = MiniVault()
    else:
        raise ResourceNotFoundError(f"The resource to load the vault from doesn't exist and live-update is not defined", varvault_filehandler_from)

    vault = VarVault(varvault_keyring, varvault_vault_name, *flags,
                     varvault_filehandler_from=varvault_filehandler_from,
                     varvault_filehandler_to=varvault_filehandler_to,
                     varvault_specific_logger=varvault_specific_logger,
                     **extra_keys)
    vault.vault.put(mini)

    if len(keys_not_in_keyring) and ignore_keys_not_in_keyring:
        vault.log(f"Vault was created from file '{varvault_filehandler_from.path}', and the file contained keys not in the keyring ({keys_not_in_keyring}), "
                  f"but you asked to ignore if this happens through setting 'ignore_keys_not_in_keyring' to True.", level=logging.WARNING)
    else:
        vault.log(f"Vault created from file '{varvault_filehandler_from.path}'.", level=logging.INFO)

    return vault


def _check_for_keys_not_in_keyring(varvault_keyring: Type[Keyring], varvault_filehandler_from: BaseFileHandler, ignore_keys_not_in_keyring: bool, *flags: VaultFlags, **extra_keys):
    vault_file_data = varvault_filehandler_from.read()

    assert isinstance(vault_file_data, dict), f"It appears we were not able to load a Dict from {varvault_filehandler_from}. Are you sure this is a valid resource? (content={vault_file_data})"
    keys_in_keyring = varvault_keyring.get_keys_in_keyring()
    keys_in_keyring.update(extra_keys)
    keys_in_file = vault_file_data.keys()
    _keys_not_in_keyring = [k for k in keys_in_file if k not in keys_in_keyring]
    assert len(_keys_not_in_keyring) == 0 or ignore_keys_not_in_keyring, \
        f"Keys found in vault-path '{varvault_filehandler_from}' that are not in the keyring, " \
        f"and you have not set to ignore keys found that are not in the keyring. " \
        f"Keys not in keyring: {_keys_not_in_keyring}"

    return _keys_not_in_keyring



import re
import inspect
import warnings
import functools

from typing import Callable


def validator(function_asserts: bool = None, function_returns_bool: bool = None, skip_source_assertions: bool = False) -> Callable:
    f"""
    Decorator that is used to register a function as a validator-function used for varvault-keys.
    A validator-function assigned to a key will be executed based on how it's registered.
    
    If no argument is passed to the function, {function_asserts} will be set to {True}.

    :param function_asserts: Used to tell varvault that the validator-function will do the validation by performing an assert
    :param function_returns_bool: Used to tell varvault that the validator-function will return a boolean that varvault will assert to be True.
    :param skip_source_assertions: Used to tell this decorator to skip source-assertions that verify that assert or return is done in the validator-function.  
    """
    assert isinstance(function_asserts, bool) or function_asserts is None, f"'function_asserts' must be a bool, or {None}"
    assert isinstance(function_returns_bool, bool) or function_returns_bool is None, f"'function_returns_bool' must be a bool, or {None}"
    if function_asserts and function_returns_bool:
        raise SyntaxError("You've set both 'function_asserts' and 'function_returns_bool' to True; 'function_asserts' takes precedence. Is this what you really want?")
    if not function_asserts and not function_returns_bool:
        function_asserts = True

    def wrap_outer(func: Callable) -> Callable:

        assert callable(func), f"Object {func} is not a callable"
        argspecs = inspect.getfullargspec(func)

        assert len(argspecs.args) == 1, f"You've defined more arguments (or no arguments) for the validator function than you are allowed to. There must be exactly one argument defined in the signature."

        source = inspect.getsource(func)
        if function_asserts:
            assert re.findall(re.compile(r".*assert .*"), source) or skip_source_assertions, "No assert-statement found in the functions source. That doesn't seem right if you are planning to do an assert in the function. Set 'skip_source_assertions=True' to skip it"
        elif function_returns_bool:
            if argspecs.annotations.get("return") != bool:
                warnings.warn("If you define the validator-function to say that it returns a bool, you should really also annotate the function to say it returns a bool.", SyntaxWarning)
            assert re.findall(re.compile(r".*return .*"), source) or skip_source_assertions, "No return-statement found in the functions source. That doesn't seem right if you are planning to return a bool that varvault is meant to assert. Set 'skip_source_assertions=True' to skip"

        @functools.wraps(func)
        def wrap_inner(keyvalue: object):
            assert callable(func), f"Function {func} is not a callable"
            if function_asserts:
                func(keyvalue)
            elif function_returns_bool:
                ret = func(keyvalue)
                assert isinstance(ret, bool), f"The return value is of type {type(ret)}, not of type {bool}, " \
                                              f"and you have said that the function will return a bool through 'function_returns_bool' being set to {True}"
                assert ret is True, f"The validator-function {func.__name__} returned False for key-value: {keyvalue}"
        return wrap_inner
    return wrap_outer
